"use strict";

let ConcentrationGame = (function() {

  /// ユーティリティ
  let U = {
    init: function(reset) {
    }
    /// 配列をn個ずつに切り出す。例：
    ///   chunk([1,2,3,4,5], 2); //=> [[1,2],[3,4],[5]]
  , chunk: function chunk(arr, n) {
      let rows = [];
      for (let i = 0, len = arr.length; i < len; i += n) {
        let row = arr.slice(i, i + n);
        rows.push(row);
      }
      return rows;
    }
  };

  /// モデル（コアとなるロジックやデータ）
  let M = {
    init: function(reset) {
      M.boardSize = 4;            // ボードの一辺の大きさ
      M.board = M.createBoard();  // ボード（4×4のマス目）を生成
    }
    /// ボードを生成する
  , createBoard: function() {
      // 1〜8までの数字をもつ配列を生成（4×4の場合）
      let n = M.boardSize;
      let nums = _.range(1, n * n / 2 + 1);
      // 自分自身を連結して、長さ16の配列を生成
      // （神経衰弱なので、同じ数字が2組ずつ必要）
      nums = nums.concat(nums);
      // ランダムに並び替え
      nums = _.shuffle(nums);
      // 数値を4個ずつ4回取り出して、4×4のマス目を生成
      let board = U.chunk(nums, n);
      return board;
    }
  , /// 指定座標の値を返す
    valueAt: function(x, y) {
      return M.board[x][y];
    }
  , /// マスの値が前回と今回で同じならtrueを返す
    challenge: function(prev, curr) {
      let v1 = M.valueAt(prev.x, prev.y);
      let v2 = M.valueAt(curr.x, curr.y);
      return v1 === v2;
    }
  };

  /// コントローラ（イベントハンドラ）
  let C = {
    init: function(reset) {
      if (! reset) {           // リセット時の重複登録を避ける
        C.registerHandlers();  // イベントハンドラを登録
      }
      V.renderBoard(M.board);  // ボードを表示
      C._previous = null;      // 前回の座標
      C._clickable = true;     // クリック可能かどうか
    }
    /// イベントハンドラを登録
  , registerHandlers: function() {
      $('#reset').on('click', C.doReset);  // リセットボタン
      $('#board').on('click', 'td', C.doFlip); // マス目
    }
    /// 全体をリセット
  , doReset: function() {
      if (confirm("Reset?")) {    // ダイアログでユーザに確認し、
        init(true);               // OKならUとMとVとCをリセット
      }
    }
    /// マス目をめくって値を表示させる
  , doFlip: function() {
      if (! C._clickable) return;   // クリック不可ならクリックを無視
      let id = $(this).attr('id');  // 左上 'c00'、右下 'c33'
      let x = +id[1], y = +id[2];   // ID属性から座標を計算
      if (V.showing(x, y)) return;  // すでに表示中ならクリックを無視
      let val = M.valueAt(x, y);    // マスの値を取得
      V.show(x, y, val);            // マスに値を表示
      //
      let curr = {x: x, y: y};      // 現在の座標
      let prev = C._previous;       // 前回の座標
      if (prev === null) {          // 1枚目のクリックなら
        C._previous = curr;         //   座標を保存して2枚目で使う
      } else {                      // 2枚目のクリックなら
        C._previous = null;         //   座標クリアして1枚目に備える
        let ok = M.challenge(prev, curr);  // 同じ値かどうか調べる
        ok ? C._whenOk(prev, curr)         // 正解したとき
           : C._whenNotOk(prev, curr);     // 不正解のとき
      }
    }
    /// 正解だったときの処理
  , _whenOk: function(prev, curr) {
      let ms = C._delay;            // 1000ミリ秒（＝1秒）
      setTimeout(function() {       // 1秒待ってから
        V.correct(prev.x, prev.y);  // 1枚目を強調表示
        V.correct(curr.x, curr.y);  // 2枚目を強調表示
      }, ms);
    }
    /// 不正解だったときの処理
  , _whenNotOk: function(prev, curr) {
      C._clickable = false;         // クリックを一時的に禁止
      let ms = C._delay;            // 1000ミリ秒（＝1秒）
      setTimeout(function() {       // 1秒待ってから
        V.hide(prev.x, prev.y);     // 1枚目の値を隠す
        V.hide(curr.x, curr.y);     // 2枚目の値を隠す
        C._clickable = true;        // 隠し終わったらクリックを再開
      }, ms);
    }
  , _delay: 1000                    // 1000ミリ秒 == 1秒
  };

  /// ビュー（表示に関する機能）
  let V = {
    init: function(reset) {
    }
    /// 空のボードを表示
  , renderBoard: function(board) {
      let context = {
        board: board
      };
      $('#tmpl-board').renderWith(context, '#board');
    }
    /// マス目に値を表示
  , show: function(x, y, val) {
      $('#c'+x+''+y).text(val).addClass('show');
    }
    /// マス目に値を表示しているならtrueを返す
  , showing: function(x, y) {
      return $('#c'+x+''+y).hasClass('show');
    }
    /// 正解したマス目を強調表示
  , correct: function(x, y) {
      $('#c'+x+''+y).addClass('correct');
    }
    /// 不正解だったマス目の値を非表示にする
  , hide: function(x, y) {
      $('#c'+x+''+y).empty().removeClass('show');
    }
  };

  /// 初期設定
  function init(reset) {
    U.init(reset);
    M.init(reset);
    V.init(reset);
    C.init(reset);
  }

  return {init: init, U: U, M: M, C: C, V: V};

})();

$(function() {
  ConcentrationGame.init();
});
