"use strict";

///
/// 使い方：
///   let r = new MartinRouter();
///   r.mount('/orders'     , funcOrders);
///   r.mount('/orders/{id}', funcOrderDetail);
///   console.log(r.route('/orders'));
///                  //=> [funcOrders, undefined]
///   console.log(r.route('/orders/123'));
///                  //=> [funcOrderDetail, '123']
///
class MartinRouter {

  constructor() {
    this._entries = [];   // パスと関数の組の配列
  }

  /// パスに対応した関数を登録する
  mount(path, func) {
    let rexp = this.compile(path);
    let x = {
      path: path        // 例: '/orders/{id}'
    , func: func        // 例: funcOrderDetail
    , rexp: rexp        // 例: new RegExp('^/orders/([^/]+)$')
    };
    this._entries.push(x);
    return this;
  }

  /// パスに対応した関数を返す（なければnullを返す）
  route(path) {
    for (let x of this._entries) {
      let m = path.match(x.rexp);        // 正規表現マッチング
      if (m) {                           // マッチしたら
        let vals = m[1];                 // パラメータの値を
        return [x.func, vals];           // 関数とともに返す
      }
    }
    return null;   // どれにもマッチしなければnullを返す
  }

  /// パスを正規表現へコンパイルする
  /// 例： '/foo.d/{id}' -> RegExp('^/foo\.d/([^/]+)$')
  compile(path) {
    let pathrexp = /(.*?)\{(\w+)\}/;
    let buf = '^';
    let m = path.match(pathrexp);
    if (m) {
      let text = m[1];    // パラメータ以外の部分は特殊文字をエスケープ
      buf += this._escape(text);
      buf += '([^/]+)';   // パラメータは '/' にマッチしてはいけない
    } else {
      buf += this._escape(path);
    }
    buf += '$';
    return new RegExp(buf);
  }

  /// 正規表現で使われる特殊文字をエスケープする
  _escape(s) {
    return s.replace(/[.*+?|^$(){}\[\]\\]/g, '\\$&');
  }

}

if (typeof module !== 'undefined') {
  module.exports = MartinRouter;
}
