"use strict";

///
/// 使い方：
///   let r = new MartinRouter();
///   r.mount('/orders'     , funcOrders);
///   r.mount('/orders/{id}', funcOrderDetail);
///   console.log(r.route('/orders'));
///                  //=> [funcOrders, {}]
///   console.log(r.route('/orders/123'));
///                  //=> [funcOrderDetail, {id: '123'}]
///
class MartinRouter {

  constructor() {
    this._entries = [];   // パスと関数の組の配列
    this._entrydict = {}; // 固定文字列用のハッシュテーブル
  }

  /// パスに対応した関数を登録する
  mount(path, func) {
    /// パスにパラメータが含まれていなければハッシュテーブルに登録
    if (! path.match(/\{/)) {
      this._entrydict[path] = {path: path, func: func};
      return this;
    }
    /// そうでなければリストに登録
    let [rexp, params] = this.compile(path);
    let x = {
      path: path        // 例: '/orders/{id}'
    , func: func        // 例: funcOrderDetail
    , rexp: rexp        // 例: new RegExp('^/orders/([^/]+)$')
    , params: params    // ex: ['id']
    };
    this._entries.push(x);
    return this;
  }

  /// パスに対応した関数を返す（なければnullを返す）
  route(path) {
    let t = this._entrydict[path];  // 最初にハッシュテーブルを検索
    if (t) {
      return [t.func, []];
    }
    for (let x of this._entries) {  // なければリストを先頭から検索
      let m = path.match(x.rexp);        // 正規表現マッチング
      if (m) {                           // マッチしたら
        let vals = m.slice(1, m.length); // すべてのパラメータの値を
        let args =
          this._combine(x.params, vals); // キーと値の組にして
        return [x.func, args];       // 関数とともに返す
      }
    }
    return null;   // どれにもマッチしなければnullを返す
  }

  /// キーの配列と値の配列から、オブジェクトを生成する
  _combine(keys, vals) {
    let obj = {};
    for (let i = 0, n = keys.length; i < n; i++) {
      obj[keys[i]] = vals[i];
    }
    return obj;
  }

  /// パスを正規表現へコンパイルする
  /// 例： '/foo.d/{id}' -> RegExp('^/foo\.d/([^/]+)$')
  compile(path) {
    let pathrexp = /(.*?)\{(\w+)\}/g;  // g: グローバルオプション
    let buf = '^', params = [], pos = 0, m;
    while ((m = pathrexp.exec(path)) != null) {
      let text = m[1];    // パラメータ以外の部分は特殊文字をエスケープ
      params.push(m[2]);               // パラメータ名を保持
      pos = m.index + m[0].length;     // '}' の位置を記録
      buf += this._escape(text);
      buf += '([^/]+)';   // パラメータは '/' にマッチしてはいけない
    }
    let rest = path.substr(pos);  // 最後の '}' より後ろの文字列
    buf += this._escape(rest);
    buf += '$';
    /// 正規表現オブジェクトと、パラメータ名の配列を返す
    return [new RegExp(buf), params];
  }

  /// 正規表現で使われる特殊文字をエスケープする
  _escape(s) {
    return s.replace(/[.*+?|^$(){}\[\]\\]/g, '\\$&');
  }

}

if (typeof module !== 'undefined') {
  module.exports = MartinRouter;
}
